@@ -29,14 +29,54 @@
 -- The following headers are automatically set by this module, and should not
 -- be added to 'requestHeaders':
 --
+-- * Cookie
+--
 -- * Content-Length
 --
+-- * Transfer-Encoding
+--
 -- Note: In previous versions, the Host header would be set by this module in
 -- all cases. Starting from 1.6.1, if a Host header is present in
 -- @requestHeaders@, it will be used in place of the header this module would
 -- have generated. This can be useful for calling a server which utilizes
 -- virtual hosting.
 --
+-- Use `cookieJar` If you want to supply cookies with your request:
+--
+-- > {-# LANGUAGE OverloadedStrings #-}
+-- > import Network.HTTP.Conduit
+-- > import Network
+-- > import Data.Time.Clock
+-- > import Data.Time.Calendar
+-- > import qualified Control.Exception as E
+-- >
+-- > past :: UTCTime
+-- > past = UTCTime (ModifiedJulianDay 56200) (secondsToDiffTime 0)
+-- >
+-- > future :: UTCTime
+-- > future = UTCTime (ModifiedJulianDay 562000) (secondsToDiffTime 0)
+-- >
+-- > cookie :: Cookie
+-- > cookie = Cookie { cookie_name = "password_hash"
+-- >                 , cookie_value = "abf472c35f8297fbcabf2911230001234fd2"
+-- >                 , cookie_expiry_time = future
+-- >                 , cookie_domain = "example.com"
+-- >                 , cookie_path = "/"
+-- >                 , cookie_creation_time = past
+-- >                 , cookie_last_access_time = past
+-- >                 , cookie_persistent = False
+-- >                 , cookie_host_only = False
+-- >                 , cookie_secure_only = False
+-- >                 , cookie_http_only = False
+-- >                 }
+-- >
+-- > main = withSocketsDo $ do
+-- >      request' <- parseUrl "http://example.com/secret-page"
+-- >      let request = request' { cookieJar = Just $ createCookieJar [cookie] }
+-- >      E.catch (withManager $ httpLbs request)
+-- >              (\(StatusCodeException s _ _) ->
+-- >                if statusCode==403 then putStrLn "login failed" else return ())
+--
 -- Any network code on Windows requires some initialization, and the network
 -- library provides withSocketsDo to perform it. Therefore, proper usage of
 -- this library will always involve calling that function at some point.  The
@@ -75,7 +115,6 @@ module Network.HTTP.Conduit
       -- * Datatypes
     , Proxy (..)
     , RequestBody (..)
-    , Response (..)
       -- ** Request
     , Request
     , def
@@ -90,11 +129,21 @@ module Network.HTTP.Conduit
     , requestBody
     , proxy
     , socksProxy
+    , hostAddress
     , rawBody
     , decompress
     , redirectCount
     , checkStatus
     , responseTimeout
+    , cookieJar
+    , getConnectionWrapper
+      -- * Response
+    , Response
+    , responseStatus
+    , responseVersion
+    , responseHeaders
+    , responseBody
+    , responseCookieJar
       -- * Manager
     , Manager
     , newManager
@@ -113,13 +162,6 @@ module Network.HTTP.Conduit
     , CookieJar
     , createCookieJar
     , destroyCookieJar
-    , updateCookieJar
-    , receiveSetCookie
-    , generateCookie
-    , insertCheckedCookie
-    , insertCookiesIntoRequest
-    , computeCookieString
-    , evictExpiredCookies
       -- * Utility functions
     , parseUrl
     , applyBasicAuth
@@ -147,19 +189,13 @@ import qualified Data.ByteString.Lazy as L
 import qualified Network.HTTP.Types as W
 import Data.Default (def)
 
-import Control.Exception.Lifted (throwIO)
+import Control.Exception.Lifted (throwIO, try, IOException, handle)
 import Control.Monad ((<=<))
 import Control.Monad.IO.Class (MonadIO (liftIO))
 import Control.Monad.Trans.Resource
-import Control.Monad.Trans.State (get, put, evalStateT)
-import Control.Monad.Trans (lift)
 
-import Control.Exception (fromException, toException)
 import qualified Data.Conduit as C
-import qualified Data.Conduit.Binary as CB
-import qualified Data.Conduit.Internal as CI
 import Data.Conduit.Blaze (builderToByteString)
-import Control.Exception.Lifted (try, IOException)
 
 import Data.Time.Clock
 
@@ -168,7 +204,8 @@ import Network.HTTP.Conduit.Response
 import Network.HTTP.Conduit.Manager
 import Network.HTTP.Conduit.ConnInfo
 import Network.HTTP.Conduit.Cookies
-import Network.HTTP.Conduit.Internal (httpRedirect)
+import Network.HTTP.Conduit.Internal (httpRedirect, applyCheckStatus)
+import Network.HTTP.Conduit.Types
 
 -- | The most low-level function for initiating an HTTP request.
 --
@@ -187,6 +224,33 @@ import Network.HTTP.Conduit.Internal (httpRedirect)
 -- You may also directly connect the returned 'C.Source'
 -- into a 'C.Sink', perhaps a file or another socket.
 --
+-- An important note: the response body returned by this function represents a
+-- live HTTP connection. As such, if you do not use the response body, an open
+-- socket will be retained until the containing @ResourceT@ block exits. If you
+-- do not need the response body, it is recommended that you explicitly shut
+-- down the connection immediately, using the pattern:
+--
+-- > responseBody res $$+- return ()
+--
+-- As a more thorough example, consider the following program. Without the
+-- explicit response body closing, the program will run out of file descriptors
+-- around the 1000th request (depending on the operating system limits).
+--
+-- > import Control.Monad          (replicateM_)
+-- > import Control.Monad.IO.Class (liftIO)
+-- > import Data.Conduit           (($$+-))
+-- > import Network                (withSocketsDo)
+-- > import Network.HTTP.Conduit
+-- >
+-- > main = withSocketsDo $ withManager $ \manager -> do
+-- >     req <- parseUrl "http://localhost/"
+-- >     mapM_ (worker manager req) [1..5000]
+-- >
+-- > worker manager req i = do
+-- >     res <- http req manager
+-- >     responseBody res $$+- return () -- The important line
+-- >     liftIO $ print (i, responseStatus res)
+--
 -- Note: Unlike previous versions, this function will perform redirects, as
 -- specified by the 'redirectCount' setting.
 http
@@ -194,40 +258,21 @@ http
     => Request m
     -> Manager
     -> m (Response (C.ResumableSource m S.ByteString))
-http req0 manager = do
-    res@(Response status _version hs body) <-
+http req0 manager = wrapIOException $ do
+    res <-
         if redirectCount req0 == 0
             then httpRaw req0 manager
-            else go (redirectCount req0) req0 def
-    case checkStatus req0 status hs of
-        Nothing -> return res
-        Just exc -> do
-            exc' <-
-                case fromException exc of
-                    Just (StatusCodeException s hdrs) -> do
-                        lbs <- body C.$$+- CB.take 1024
-                        return $ toException $ StatusCodeException s $ hdrs ++
-                            [("X-Response-Body-Start", S.concat $ L.toChunks lbs)]
-                    _ -> do
-                        let CI.ResumableSource _ final = body
-                        final
-                        return exc
-            liftIO $ throwIO exc'
+            else go (redirectCount req0) req0
+    maybe (return res) throwIO =<< applyCheckStatus (checkStatus req0) res
   where
-    go count req''' cookie_jar''' = (`evalStateT` cookie_jar''') $
-      httpRedirect
+    go count req' = httpRedirect
       count
-      (\req'' -> do
-        cookie_jar'' <- get
-        now <- liftIO getCurrentTime
-        let (req', cookie_jar') = insertCookiesIntoRequest req'' (evictExpiredCookies cookie_jar'' now) now
-        res <- lift $ httpRaw req' manager
-        let (cookie_jar, _) = updateCookieJar res req' now cookie_jar'
-        put cookie_jar
-        let mreq = getRedirectedRequest req' (responseHeaders res) (W.statusCode (responseStatus res))
+      (\req -> do
+        res <- httpRaw req manager
+        let mreq = getRedirectedRequest req (responseHeaders res) (responseCookieJar res) (W.statusCode (responseStatus res))
         return (res, mreq))
-      lift
-      req'''
+      id
+      req'
 
 -- | Get a 'Response' without any redirect following.
 httpRaw
@@ -235,8 +280,13 @@ httpRaw
      => Request m
      -> Manager
      -> m (Response (C.ResumableSource m S.ByteString))
-httpRaw req m = do
-    (connRelease, ci, isManaged) <- getConnectionWrapper
+httpRaw req' m = do
+    (req, cookie_jar') <- case cookieJar req' of
+        Just cj -> do
+            now <- liftIO getCurrentTime
+            return $ insertCookiesIntoRequest req' (evictExpiredCookies cj now) now
+        Nothing -> return (req', def)
+    (timeout', (connRelease, ci, isManaged)) <- getConnectionWrapper
         req
         (responseTimeout req)
         (failedConnectionException req)
@@ -249,7 +299,7 @@ httpRaw req m = do
     -- exceptions in both.
     ex <- try' $ do
         requestBuilder req C.$$ builderToByteString C.=$ connSink ci
-        getResponse connRelease req src
+        getResponse connRelease timeout' req src
 
     case (ex, isManaged) of
         -- Connection was reused, and might be been closed. Try again
@@ -260,7 +310,12 @@ httpRaw req m = do
         (Left e, Fresh) -> liftIO $ throwIO e
         -- Everything went ok, so the connection is good. If any exceptions get
         -- thrown in the response body, just throw them as normal.
-        (Right x, _) -> return x
+        (Right res, _) -> case cookieJar req' of
+            Just _ -> do
+                now' <- liftIO getCurrentTime
+                let (cookie_jar, _) = updateCookieJar res req now' cookie_jar'
+                return $ res {responseCookieJar = cookie_jar}
+            Nothing -> return res
   where
     try' :: MonadBaseControl IO m => m a -> m (Either IOException a)
     try' = try
@@ -283,7 +338,10 @@ httpRaw req m = do
 -- Note: Unlike previous versions, this function will perform redirects, as
 -- specified by the 'redirectCount' setting.
 httpLbs :: (MonadBaseControl IO m, MonadResource m) => Request m -> Manager -> m (Response L.ByteString)
-httpLbs r = lbsResponse <=< http r
+httpLbs r = wrapIOException . (lbsResponse <=< http r)
+
+wrapIOException :: MonadBaseControl IO m => m a -> m a
+wrapIOException = handle $ throwIO . InternalIOException
 
 -- | Download the specified URL, following any redirects, and
 -- return the response body.
