@@ -7,6 +7,7 @@ module Network.HTTP.Conduit.Manager
     ( Manager
     , ManagerSettings (..)
     , ConnKey (..)
+    , ConnHost (..)
     , newManager
     , closeManager
     , getConn
@@ -37,7 +38,7 @@ import qualified Data.Text as T
 
 import Control.Monad.Trans.Control (MonadBaseControl)
 import Control.Monad.IO.Class (MonadIO, liftIO)
-import Control.Exception (mask_, SomeException, catch)
+import Control.Exception (mask_, SomeException, catch, throwIO, fromException)
 import Control.Monad.Trans.Resource
     ( ResourceT, runResourceT, MonadResource
     , MonadThrow, MonadUnsafeIO
@@ -47,8 +48,7 @@ import Control.Concurrent (forkIO, threadDelay)
 import Data.Time (UTCTime (..), Day (..), DiffTime, getCurrentTime, addUTCTime)
 import Control.DeepSeq (deepseq)
 
-import Network (connectTo, PortID (PortNumber), HostName)
-import Network.Socket (socketToHandle)
+import qualified Network.Socket as NS
 import Data.Certificate.X509 (X509, encodeCertificate)
 import Data.CertificateStore (CertificateStore)
 import System.Certificate.X509 (getSystemCertificateStore)
@@ -59,11 +59,12 @@ import Network.TLS.Extra (certificateVerifyChain, certificateVerifyDomain)
 import Network.HTTP.Conduit.ConnInfo
 import Network.HTTP.Conduit.Types
 import Network.HTTP.Conduit.Util (hGetSome)
-import Network.HTTP.Conduit.Parser (parserHeadersFromByteString)
-import Network.Socks5 (SocksConf, socksConnectWith)
+import Network.HTTP.Conduit.Parser (sinkHeaders)
+import Network.Socks5 (SocksConf)
 import Data.Default
 import Data.Maybe (mapMaybe)
 import System.IO (Handle)
+import Data.Conduit (($$), yield, runException)
 
 -- | Settings for a @Manager@. Please use the 'def' function and then modify
 -- individual settings.
@@ -110,9 +111,15 @@ data NonEmptyList a =
     One !a !UTCTime |
     Cons !a !Int !UTCTime !(NonEmptyList a)
 
+-- | Hostname or resolved host address.
+data ConnHost =
+    HostName !Text |
+    HostAddress !NS.HostAddress
+    deriving (Eq, Show, Ord)
+
 -- | @ConnKey@ consists of a hostname, a port and a @Bool@
 -- specifying whether to use SSL.
-data ConnKey = ConnKey !Text !Int !Bool
+data ConnKey = ConnKey !ConnHost !Int !Bool
     deriving (Eq, Show, Ord)
 
 takeSocket :: Manager -> ConnKey -> IO (Maybe ConnInfo)
@@ -294,70 +301,66 @@ nonEmptyMapM_ f (One x _) = f x
 nonEmptyMapM_ f (Cons x _ _ l) = f x >> nonEmptyMapM_ f l
 
 getSocketConn
-    :: MonadResource m
-    => Manager
+    :: Maybe NS.HostAddress
     -> String
     -> Int
     -> Maybe SocksConf -- ^ optional socks proxy
-    -> m (ConnRelease m, ConnInfo, ManagedConn)
-getSocketConn man host' port' socksProxy' =
-    getManagedConn man (ConnKey (T.pack host') port' False) $
-        getSocket host' port' socksProxy' >>= socketConn desc
+    -> IO ConnInfo
+getSocketConn hostAddress' host' port' socksProxy' =
+    getSocket hostAddress' host' port' socksProxy' >>= socketConn desc
   where
     desc = socketDesc host' port' "unsecured"
 
 socketDesc :: String -> Int -> String -> String
 socketDesc h p t = unwords [h, show p, t]
 
-getSslConn :: MonadResource m
-            => ([X509] -> IO CertificateUsage)
+getSslConn ::  ([X509] -> IO CertificateUsage)
             -> [(X509, Maybe PrivateKey)]
-            -> Manager
+            -> Maybe NS.HostAddress
             -> String -- ^ host
             -> Int -- ^ port
             -> Maybe SocksConf -- ^ optional socks proxy
-            -> m (ConnRelease m, ConnInfo, ManagedConn)
-getSslConn checkCert clientCerts man host' port' socksProxy' =
-    getManagedConn man (ConnKey (T.pack host') port' True) $
-        (connectionTo host' (PortNumber $ fromIntegral port') socksProxy' >>= sslClientConn desc host' checkCert clientCerts)
+            -> IO ConnInfo
+getSslConn checkCert clientCerts hostAddress' host' port' socksProxy' =
+    connectionTo hostAddress' host' port' socksProxy' >>= sslClientConn desc host' checkCert clientCerts
   where
     desc = socketDesc host' port' "secured"
 
 getSslProxyConn
-            :: MonadResource m
-            => ([X509] -> IO CertificateUsage)
+            :: ([X509] -> IO CertificateUsage)
             -> [(X509, Maybe PrivateKey)]
             -> S8.ByteString -- ^ Target host
             -> Int -- ^ Target port
-            -> Manager
+            -> Maybe NS.HostAddress
             -> String -- ^ Proxy host
             -> Int -- ^ Proxy port
             -> Maybe SocksConf -- ^ optional SOCKS proxy
-            -> m (ConnRelease m, ConnInfo, ManagedConn)
-getSslProxyConn checkCert clientCerts thost tport man phost pport socksProxy' =
-    getManagedConn man (ConnKey (T.pack phost) pport True) $
-        doConnect >>= sslClientConn desc phost checkCert clientCerts
+            -> IO ConnInfo
+getSslProxyConn checkCert clientCerts thost tport phostAddr phost pport socksProxy' =
+    doConnect >>= sslClientConn desc phost checkCert clientCerts
   where
     desc = socketDesc phost pport "secured-proxy"
     doConnect = do
-        h <- connectionTo phost (PortNumber $ fromIntegral pport) socksProxy'
+        h <- connectionTo phostAddr phost pport socksProxy'
         L.hPutStr h $ Blaze.toLazyByteString connectRequest
         hFlush h
         r <- hGetSome h 2048
-        res <- parserHeadersFromByteString r
-        case res of
+        case runException $ yield r $$ sinkHeaders of
             Right ((_, 200, _), _) -> return h
-            Right ((_, _, msg), _) -> hClose h >> proxyError (S8.unpack msg)
-            Left s -> hClose h >> proxyError s
+            Right ((_, _, msg), _) -> hClose h >> proxyError (Left msg)
+            Left s -> do
+                hClose h
+                proxyError $
+                    case fromException s of
+                        Just he -> Right he
+                        Nothing -> Left $ S8.pack $ show s
 
     connectRequest =
         Blaze.fromByteString "CONNECT "
             `mappend` Blaze.fromByteString thost
             `mappend` Blaze.fromByteString (S8.pack (':' : show tport))
             `mappend` Blaze.fromByteString " HTTP/1.1\r\n\r\n"
-    proxyError s =
-        error $ "Proxy failed to CONNECT to '"
-                ++ S8.unpack thost ++ ":" ++ show tport ++ "' : " ++ s
+    proxyError s = throwIO $ ProxyConnectException thost tport s
 
 -- | This function needs to acquire a @ConnInfo@- either from the @Manager@ or
 -- via I\/O, and register it with the @ResourceT@ so it is guaranteed to be
@@ -423,10 +426,15 @@ getConn :: MonadResource m
         -> Manager
         -> m (ConnRelease m, ConnInfo, ManagedConn)
 getConn req m =
-    go m connhost connport (socksProxy req)
+    getManagedConn m (ConnKey connKeyHost connport (secure req)) $
+        go connaddr connhost connport (socksProxy req)
   where
     h = host req
     (useProxy, connhost, connport) = getConnDest req
+    (connaddr, connKeyHost) =
+        case (hostAddress req, useProxy, socksProxy req) of
+            (Just ha, False, Nothing) -> (Just ha, HostAddress ha)
+            _ -> (Nothing, HostName $ T.pack connhost)
     go =
         case (secure req, useProxy) of
             (False, _) -> getSocketConn
@@ -471,7 +479,6 @@ checkCerts man host' certs = do
                 Nothing -> Map.singleton encoded expire
                 Just m -> Map.insert encoded expire m
 
-connectionTo :: HostName -> PortID -> Maybe SocksConf -> IO Handle
-connectionTo host' port' Nothing = connectTo host' port'
-connectionTo host' port' (Just socksConf) =
-	socksConnectWith socksConf host' port' >>= flip socketToHandle ReadWriteMode
+connectionTo :: Maybe NS.HostAddress -> NS.HostName -> Int -> Maybe SocksConf -> IO Handle
+connectionTo hostAddress' host' port' socksConf' =
+    getSocket hostAddress' host' port' socksConf' >>= flip NS.socketToHandle ReadWriteMode
