@@ -1,6 +1,7 @@
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE CPP #-}
 module Network.HTTP.Conduit.Response
     ( Response (..)
     , getRedirectedRequest
@@ -8,9 +9,8 @@ module Network.HTTP.Conduit.Response
     , lbsResponse
     ) where
 
-import Control.Applicative ((<$>), (<*>))
 import Control.Arrow (first)
-import Control.Monad (liftM, unless, when)
+import Control.Monad (liftM)
 
 import Control.Exception (throwIO)
 import Control.Monad.IO.Class (liftIO)
@@ -21,6 +21,8 @@ import qualified Data.ByteString.Lazy as L
 
 import qualified Data.CaseInsensitive as CI
 
+import Data.Default (def)
+
 import Data.Conduit hiding (Conduit)
 import Data.Conduit.Internal (ResumableSource (..), Pipe (..))
 import qualified Data.Conduit.Zlib as CZ
@@ -30,16 +32,20 @@ import qualified Data.Conduit.List as CL
 import qualified Network.HTTP.Types as W
 import Network.URI (parseURIReference)
 
-import Network.HTTP.Conduit.Types (Response (..))
+import Network.HTTP.Conduit.Types (Response (..), CookieJar)
 
 import Network.HTTP.Conduit.Manager
 import Network.HTTP.Conduit.Request
 import Network.HTTP.Conduit.Util
 import Network.HTTP.Conduit.Chunk
+import Network.HTTP.Conduit.Parser (sinkHeaders)
 
 import Data.Void (Void, absurd)
 
 import System.Timeout.Lifted (timeout)
+#if MIN_VERSION_conduit(1, 0, 0)
+import Data.Conduit.Internal (ConduitM (..))
+#endif
 
 -- | If a request is a redirection (status code 3xx) this function will create
 -- a new request from the old request, the server headers returned with the
@@ -51,14 +57,21 @@ import System.Timeout.Lifted (timeout)
 -- specific request, that user has to re-implement the redirect-following logic
 -- themselves. An example of that might look like this:
 --
--- > myHttp req man = E.catch (runResourceT $ http req' man >> return [req'])
--- >                    (\ (StatusCodeException status headers) -> do
--- >                        l <- myHttp (fromJust $ nextRequest status headers) man
--- >                        return $ req' : l)
--- >     where req' = req { redirectCount = 0 }
--- >           nextRequest status headers = getRedirectedRequest req' headers $ W.statusCode status
-getRedirectedRequest :: Request m -> W.ResponseHeaders -> Int -> Maybe (Request m)
-getRedirectedRequest req hs code
+-- > myHttp req man = do
+-- >    (res, redirectRequests) <- (`runStateT` []) $
+-- >         'httpRedirect'
+-- >             9000
+-- >             (\req' -> do
+-- >                res <- http req'{redirectCount=0} man
+-- >                modify (\rqs -> req' : rqs)
+-- >                return (res, getRedirectedRequest req' (responseHeaders res) (responseCookieJar res) (W.statusCode (responseStatus res))
+-- >                )
+-- >             'lift'
+-- >             req
+-- >    applyCheckStatus (checkStatus req) res
+-- >    return redirectRequests
+getRedirectedRequest :: Request m -> W.ResponseHeaders -> CookieJar -> Int -> Maybe (Request m)
+getRedirectedRequest req hs cookie_jar code
     | 300 <= code && code < 400 = do
         l' <- lookup "location" hs
         req' <- setUriRelative req =<< parseURIReference (S8.unpack l')
@@ -70,9 +83,12 @@ getRedirectedRequest req hs code
                 then req'
                     { method = "GET"
                     , requestBody = RequestBodyBS ""
+                    , cookieJar = cookie_jar'
                     }
-                else req'
+                else req' {cookieJar = cookie_jar'}
     | otherwise = Nothing
+  where
+    cookie_jar' = fmap (const cookie_jar) $ cookieJar req
 
 -- | Convert a 'Response' that has a 'Source' body to one with a lazy
 -- 'L.ByteString' body.
@@ -103,19 +119,25 @@ checkHeaderLength len (Leftover p i) = Leftover (checkHeaderLength (len + S.leng
 
 getResponse :: (MonadResource m, MonadBaseControl IO m)
             => ConnRelease m
+            -> Maybe Int
             -> Request m
             -> Source m S8.ByteString
             -> m (Response (ResumableSource m S8.ByteString))
-getResponse connRelease req@(Request {..}) src1 = do
+getResponse connRelease timeout'' req@(Request {..}) src1 = do
     let timeout' =
-            case responseTimeout of
+            case timeout'' of
                 Nothing -> id
                 Just useconds -> \ma -> do
                     x <- timeout useconds ma
                     case x of
                         Nothing -> liftIO $ throwIO ResponseTimeout
                         Just y -> return y
-    (src2, ((vbs, sc, sm), hs)) <- timeout' $ src1 $$+ checkHeaderLength 4096 sinkHeaders'
+    (src2, ((vbs, sc, sm), hs)) <- timeout' $ src1 $$+
+#if MIN_VERSION_conduit(1, 0, 0)
+        ConduitM (checkHeaderLength 4096 $ unConduitM sinkHeaders)
+#else
+        (checkHeaderLength 4096 sinkHeaders)
+#endif
     let version = if vbs == "1.1" then W.http11 else W.http10
     let s = W.Status sc sm
     let hs' = map (first CI.mk) hs
@@ -146,77 +168,7 @@ getResponse connRelease req@(Request {..}) src1 = do
                             else src3
                 return $ addCleanup' cleanup src4
 
-    return $ Response s version hs' body
+    return $ Response s version hs' body def
   where
     fmapResume f (ResumableSource src m) = ResumableSource (f src) m
     addCleanup' f (ResumableSource src m) = ResumableSource (addCleanup f src) (m >> f False)
-
--- | New version of @sinkHeaders@ that doesn't use attoparsec. Should create
--- more meaningful exceptions.
---
--- Since 1.8.7
-sinkHeaders' :: (MonadThrow m, MonadResource m) => Sink S.ByteString m (Status, [Header])
-sinkHeaders' = do
-    status <- getStatusLine
-    headers <- parseHeaders id
-    return (status, headers)
-  where
-    getStatusLine = do
-        status@(_, code, _) <- sinkLine >>= parseStatus
-        if code == 100
-            then newline ExpectedBlankAfter100Continue >> getStatusLine
-            else return status
-
-    newline exc = do
-        line <- sinkLine
-        unless (S.null line) $ monadThrow exc
-
-    sinkLine = do
-        bs <- fmap (killCR . S.concat) $ CB.takeWhile (/= charLF) =$ CL.consume
-        CB.drop 1
-        return bs
-    charLF = 10
-    charCR = 13
-    charSpace = 32
-    charColon = 58
-    killCR bs
-        | S.null bs = bs
-        | S.last bs == charCR = S.init bs
-        | otherwise = bs
-
-    parseStatus :: MonadThrow m => S.ByteString -> m Status
-    parseStatus bs = do
-        let (ver, bs2) = S.breakByte charSpace bs
-            (code, bs3) = S.breakByte charSpace $ S.dropWhile (== charSpace) bs2
-            msg = S.dropWhile (== charSpace) bs3
-        case (,) <$> parseVersion ver <*> parseCode code of
-            Just (ver', code') -> return (ver', code', msg)
-            _ -> monadThrow $ InvalidStatusLine bs
-
-    stripPrefixBS x y
-        | x `S.isPrefixOf` y = Just $ S.drop (S.length x) y
-        | otherwise = Nothing
-    parseVersion = stripPrefixBS "HTTP/"
-    parseCode bs =
-        case S8.readInt bs of
-            Just (i, "") -> Just i
-            _ -> Nothing
-
-    parseHeaders front = do
-        line <- sinkLine
-        if S.null line
-            then return $ front []
-            else do
-                header <- parseHeader line
-                parseHeaders $ front . (header:)
-
-    parseHeader :: MonadThrow m => S.ByteString -> m Header
-    parseHeader bs = do
-        let (key, bs2) = S.breakByte charColon bs
-        when (S.null bs2) $ monadThrow $ InvalidHeader bs
-        return (strip key, strip $ S.drop 1 bs2)
-
-    strip = S.dropWhile (== charSpace) . fst . S.spanEnd (== charSpace)
-
-type Header = (S.ByteString, S.ByteString)
-type Status = (S.ByteString, Int, S.ByteString)
